
#include "x.h"
#include "xm.h"
#include "xpex.h"
#include "off.h"
#include "defs.h"
#include "externs.h"
#include "dialog.h"
#include "resources.h"          /* fallback resources needed for
                                   reasonable standalone operation */
#include <X11/Xmu/Editres.h>

#ifdef XPM
#include <X11/xpm.h>
#endif

#include "unix.h"

XtAppContext app_context;

Display *display;

Widget toplevel, w_canvas;
Widget w_control, w_control_video, w_control_material_predef;
Widget w_material_predef, w_video, w_wireframe, w_cull, w_edge, w_shadow;
String XwdCommand;
String PrintCommand;
Pixmap groundPixmap;

static void quit();
static XtActionsRec top_actions[] = {
    { "quit", quit },
};

Atom wm_delete_window;

/* program line options */
static XrmOptionDescRec options[] =
{
    {"-xwdcommand", "xwdcommand", XrmoptionSepArg, NULL},
    {"-printcommand", "printcommand", XrmoptionSepArg, NULL},
};


/* new resources defined */
static XtResource resources[] =
{
    { "xwdcommand", "XwdCommand",
    XtRString, sizeof (char *), (Cardinal) &XwdCommand,
        XtRString, "xwd -frame"
    },
    { "printcommand", "PrintCommand",
    XtRString, sizeof (char *), (Cardinal) &PrintCommand,
    /*  XtRString, "xwd -frame -out /tmp/temp1 ; \
                xpr -rv -output /tmp/temp2 -device ps /tmp/temp1 ; \
                lpr /tmp/temp2" */
        XtRString, "xwd2ps"
    },
};


static void quit (w, event, params, num_params)
Widget w;
XEvent *event;
String *params;
Cardinal *num_params;
{
    extern int using_drawable;
    extern Window window;
    extern OFFObjDesc *Obj;

    if (event->type == ClientMessage &&
        event->xclient.data.l[0] != wm_delete_window) {
        XBell (XtDisplay(w), 0);
        return;
    }
    if( Obj ) OFFFreeObj( Obj );
    if( using_drawable == BUFFER )
        XmbufDestroyBuffers( display, window );
    XtDestroyApplicationContext(app_context);
    exit (0);
}


void client_msg( shell, parm, event, flag )
Widget shell;
caddr_t * parm;
XEvent * event;
int * flag;
{
    if( event->type == ClientMessage ) {
        XClientMessageEvent * cm = (XClientMessageEvent *) event;
        Atom atom;

        atom = XmInternAtom( XtDisplay(toplevel), "_MOTIF_WM_OFFSET", FALSE );

        if ( cm->data.l[0] == atom )
            fprintf( stderr, "Decoration offsets: %dx%d\n",cm->data.l[1], cm->data.l[2]);
        else {
            char * cp = XGetAtomName( XtDisplay(toplevel), cm->data.l[0] );
            fprintf( stderr, "Some other message, atom is %s\n", cp );
            XFree( cp );
        }
    }
}


void initXtAppTop( argc, argv )
unsigned int argc;
char **argv;
{
    void _XEditResCheckMessages();
    extern int PrintDefaultError();

    toplevel = XtAppInitialize( &app_context, "ViewObject",
        options, XtNumber( options ),
        (int *)&argc, argv,
        fallback_resources,
        NULL, 0 );

    XtAddEventHandler( toplevel, (EventMask)0, True,
        _XEditResCheckMessages, NULL );
#if( 0 ) 
    XtAddEventHandler( toplevel, (EventMask)0, True,
        (XtEventHandler)client_msg, NULL );
#endif
    XtGetApplicationResources( toplevel, (XtPointer) NULL,
        resources, XtNumber( resources ),
        NULL, 0 );

    display = XtDisplay( toplevel );
    XSetErrorHandler( PrintDefaultError );
#ifdef XDEBUG
    XSynchronize( display, True );
#endif
}


int
MakeMenuButton( Bar, Pane, Title, ResName, Mnwm )
Widget Bar;
Widget Pane;
char *Title;
char *ResName;
char *Mnwm;
{
    Widget Syscascade;
    Arg args[3];

    XtSetArg( args[0], XmNsubMenuId, Pane );
    XtSetArg( args[1], XmNlabelString, XmStringCreateSimple(Title) );
    XtSetArg( args[2], XmNmnemonic, Mnwm );
    Syscascade = XmCreateCascadeButton( Bar, ResName, args, 3 );
    XtManageChild( Syscascade );
    return 0;
}


void changeENV( w, client_data, call_data )
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    extern char **environment;
    char *objs = (char *)client_data;
    int i=0;

    while( environment[i] != (char *)NULL ) {
        if( !strncmp( environment[i], "OBJ_LIST", 8 ) ) {
            environment[i][strlen(environment[i])-1] = objs[3];
            environment[i][strlen(environment[i])-2] = objs[2];
            environment[i][strlen(environment[i])-3] = objs[1];
            environment[i][strlen(environment[i])-4] = objs[0];
            return;
        }
        i++;
    }
}


void createTopForm()
{
    Arg args[10];
    register int n=0;
    Widget top_pane, c_name;
    Widget c_box;
    Widget menu_bar, menu_pane1, menu_pane2,
           quitObj, saveObj, loadObj, createObj, loadFile;
    Widget quadrics, functions, blob, lSystems, revolution, extrusion,
           simple, platonic, polyhedra, fractals, shells;
    extern void createOFFObject();

    top_pane = XtVaCreateManagedWidget( "top_pane", xmPanedWindowWidgetClass,
        toplevel,
        XmNresizePolicy, XmNONE,
        NULL );

    c_name = XtVaCreateManagedWidget( "c_name", xmLabelWidgetClass,
        top_pane,
        XmNlabelString, XmStringCreateSimple( "Select OFF Object" ),
        NULL );

    c_box = XtVaCreateManagedWidget( "c_box", xmRowColumnWidgetClass,
        top_pane,
        NULL );

    menu_bar   = XmCreateMenuBar( c_box, "menu_bar", args, n );

    menu_pane1 = XmCreatePulldownMenu( menu_bar, "menu_pane1", args, n );
    createObj  = XmCreatePulldownMenu( menu_pane1, "createObj", args, n );

    menu_pane2 = XmCreatePulldownMenu( menu_bar, "menu_pane2", args, n );

    simple = XtVaCreateManagedWidget( "Simple",
        xmPushButtonWidgetClass, createObj, NULL );
    XtAddCallback( simple, XmNactivateCallback,
        (XtCallbackProc) changeENV, (XtPointer)"Smpl" );
    XtAddCallback( simple, XmNactivateCallback,
        (XtCallbackProc) createOFFObject, (XtPointer)NULL );

    quadrics = XtVaCreateManagedWidget( "Quadrics",
        xmPushButtonWidgetClass, createObj, NULL );
    XtAddCallback( quadrics, XmNactivateCallback,
        (XtCallbackProc) changeENV, (XtPointer)"Quad" );
    XtAddCallback( quadrics, XmNactivateCallback,
        (XtCallbackProc) createOFFObject, (XtPointer)NULL );

    functions = XtVaCreateManagedWidget( "Functions",
        xmPushButtonWidgetClass, createObj, NULL );
    XtAddCallback( functions, XmNactivateCallback,
        (XtCallbackProc) changeENV, (XtPointer)"Func" );
    XtAddCallback( functions, XmNactivateCallback,
        (XtCallbackProc) createOFFObject, (XtPointer)NULL );

    blob = XtVaCreateManagedWidget( "Blobs",
        xmPushButtonWidgetClass, createObj, NULL );
    XtSetSensitive( blob, False );

    revolution = XtVaCreateManagedWidget( "Object Revolution",
        xmPushButtonWidgetClass, createObj, NULL );
    XtSetSensitive( revolution, False );

    extrusion = XtVaCreateManagedWidget( "Object Extrusion",
        xmPushButtonWidgetClass, createObj, NULL );
    XtSetSensitive( extrusion, False );

    platonic = XtVaCreateManagedWidget( "Platonic Solids",
        xmPushButtonWidgetClass, createObj, NULL );
    XtAddCallback( platonic, XmNactivateCallback,
        (XtCallbackProc) changeENV, (XtPointer)"Plnc" );
    XtAddCallback( platonic, XmNactivateCallback,
        (XtCallbackProc) createOFFObject, (XtPointer)NULL );

    polyhedra = XtVaCreateManagedWidget( "Polyhedra",
        xmPushButtonWidgetClass, createObj, NULL );
    XtAddCallback( polyhedra, XmNactivateCallback,
        (XtCallbackProc) changeENV, (XtPointer)"Poly" );
    XtAddCallback( polyhedra, XmNactivateCallback,
        (XtCallbackProc) createOFFObject, (XtPointer)NULL );

    fractals = XtVaCreateManagedWidget( "Fractals",
        xmPushButtonWidgetClass, createObj, NULL );
    XtAddCallback( fractals, XmNactivateCallback,
        (XtCallbackProc) changeENV, (XtPointer)"Frac" );
    XtAddCallback( fractals, XmNactivateCallback,
        (XtCallbackProc) createOFFObject, (XtPointer)NULL );

    lSystems = XtVaCreateManagedWidget( "L-Systems",
        xmPushButtonWidgetClass, createObj, NULL );
    XtSetSensitive( lSystems, False );

    shells = XtVaCreateManagedWidget( "Sea Shells",
        xmPushButtonWidgetClass, createObj, NULL );
    XtSetSensitive( shells, False );

    MakeMenuButton( menu_pane1, createObj, "Create" , "CreateMenuButt",
        (char*)'C');

    loadObj = XtVaCreateManagedWidget( "Load", xmPushButtonWidgetClass,
        menu_pane1,
        XmNshadowType, XmSHADOW_ETCHED_IN,
        NULL );
    XtAddCallback( loadObj, XmNactivateCallback,
        (XtCallbackProc) loadOFFObject, (XtPointer)NULL );

    saveObj = XtVaCreateManagedWidget( "Save", xmPushButtonWidgetClass,
        menu_pane1,
        XmNshadowType, XmSHADOW_ETCHED_IN,
        NULL );
    XtAddCallback( saveObj, XmNactivateCallback,
        (XtCallbackProc) saveOFFObject, (XtPointer)NULL );

    quitObj = XtVaCreateManagedWidget( "Quit", xmPushButtonWidgetClass,
        menu_pane1,
        XmNshadowType, XmSHADOW_ETCHED_IN,
        NULL );
    XtAddCallback( quitObj, XmNactivateCallback,
        (XtCallbackProc) quitOFFObject, (XtPointer)NULL );

    loadFile = XtVaCreateManagedWidget( "Load File", xmPushButtonWidgetClass,
        menu_pane2,
        XmNshadowType, XmSHADOW_ETCHED_IN,
        NULL );
    XtAddCallback( loadFile, XmNactivateCallback,
        (XtCallbackProc) loadOFFFile, (XtPointer)NULL );
    XtSetSensitive( loadFile, False );

    MakeMenuButton( menu_bar, menu_pane1, "Object   " , "ObjectMenuButt",
        (char*)'O');
    MakeMenuButton( menu_bar, menu_pane2, "File     " , "FileMenuButt",
        (char*)'F');

    XtManageChild( menu_bar );
    XtRealizeWidget( toplevel );

    /*
     * This is a hack so that f.delete will do something useful in this
     * single-window application.
     */
    XtAppAddActions( app_context, top_actions, XtNumber(top_actions) );
    XtOverrideTranslations( toplevel,
           XtParseTranslationTable( "<Message>WM_PROTOCOLS: quit()" ) );
    wm_delete_window = XInternAtom( XtDisplay(toplevel), "WM_DELETE_WINDOW",
                                    False );
    (void) XSetWMProtocols( XtDisplay(toplevel), XtWindow(toplevel),
                            &wm_delete_window, 1 );
}


void loadOFFObject( w, client_data, call_data )
Widget w;
XtPointer client_data;
XmAnyCallbackStruct *call_data;
{
    char *filename;
    extern char       *FileName;
    extern OFFObjDesc *Obj;
    extern Dialog dialog_f;
    extern char * pattern;
    Position x, y;

    FileName = (char *)malloc(10);
    FileName = "  ";

    switch( call_data->reason ) {
    case XmCR_ACTIVATE:
        RetryLoad:
        filename = XtNewString( FileName );
        strcpy( pattern, "geom" );
        if( PopupFDialog( dialog_f, "Selection", filename, &FileName )==Okay ) {
            int start, last;
            int l;
            if( ( start = strbpos( FileName, '/' ) ) > 0 ) {
                int i;
                last = strlen( FileName );
                for( i=0; i<last-start; i++ )
                    FileName[i] = FileName[start+i+1];
                FileName[last-start] = '\0';
            }
            l = strlen( FileName );
            if( ((FileName[l-1] != 'f') || (FileName[l-2] != 'f') || 
                 (FileName[l-3] != 'o') || (FileName[l-4] != 'a'))
                &&
                ((FileName[l-1] != 'm') || (FileName[l-2] != 'o') || 
                 (FileName[l-3] != 'e') || (FileName[l-4] != 'g')) ) {
                warning( XtParent( w ),
                    "Extension \".geom\" OR \".bgeom\" is missing\0" );
                goto RetryLoad;
            }
            processFile( FileName );
            createControlPanel();
            OFFReadObj( Obj );
            initPex( Obj );
        }
        break;
    }
}


void loadOFFFile( w, client_data, call_data )
Widget w;
XtPointer client_data;
XmAnyCallbackStruct *call_data;
{
}


void saveOFFObject( w, client_data, call_data )
Widget w;
XtPointer client_data;
XmAnyCallbackStruct *call_data;
{
    char   message[80];
    String filename;
    extern Dialog dialog_s, dialog_e;
    extern char *FileName;
    extern int smth_changed;

    if( smth_changed ) {
        RetryQuit:
        filename = XtNewString( FileName );
        sprintf( message, "Save file \"%s\" before quitting?", filename );
        if( PopupDialog( dialog_s, message, filename, &filename ) == Yes ) {
            int l = strlen(filename);
            if( (filename[l-1] != 'f') ||
                (filename[l-2] != 'f') ||
                (filename[l-3] != 'o') ||
                (filename[l-4] != 'a') ) {
                warning( XtParent( w ),
                    "Extension \".aoff\" is missing \0" );
                goto RetryQuit;
            }
            if( ! WriteAOFF( filename ) )
            {
                sprintf( message, "Can't write file: %s", filename );
                if( PopupDialog( dialog_e, message, NULL, NULL ) == Retry )
                    goto RetryQuit;
            }
        }
    }
    else
        warning( XtParent( w ), "Nothing has changed \0" );
    smth_changed = 0;
}


void quitOFFObject( w, client_data, call_data )
Widget w;
XtPointer client_data;
XmAnyCallbackStruct *call_data;
{
    extern int using_drawable;
    extern Window window;
    extern OFFObjDesc *Obj;

    switch( call_data->reason ) {
        case XmCR_ACTIVATE:
            if( Obj ) OFFFreeObj( Obj );
            if( using_drawable == BUFFER )
                XmbufDestroyBuffers( display, window );
            XtDestroyApplicationContext(app_context);
            (void) exit (0);
        case XmCR_ARM:
        case XmCR_DISARM:
            break;
    }
}


#ifdef XPM
void errorMessageXPM(ErrorStatus, tag)
int ErrorStatus;
char *tag;
{
    char *error = NULL;
    char *warning = NULL;

    switch (ErrorStatus) {
    case XpmSuccess:
        return;
    case XpmColorError:
        warning = "Could not parse or alloc requested color";
        break;
    case XpmOpenFailed:
        error = "Cannot open file";
        break;
    case XpmFileInvalid:
        error = "Invalid XPM file";
        break;
    case XpmNoMemory:
        error = "Not enough memory";
        break;
    case XpmColorFailed:
        error = "Failed to parse or alloc some color";
        break;
    }

    if (warning)
        printf("%s Xpm Warning: %s.\n", tag, warning);

    if (error) {
        printf("%s Xpm Error: %s.\n", tag, error);
/*      exit( -1 ); */
    }
}
#endif

void errorMessageXBM(ErrorStatus, tag)
int ErrorStatus;
char *tag;
{
    char *error = NULL;
    char *warning = NULL;

    switch (ErrorStatus) {
    case BitmapSuccess:
        return;
    case BitmapOpenFailed:
        error = "Cannot open file";
        break;
    case BitmapFileInvalid:
        error = "Invalid XBM file";
        break;
    case BitmapNoMemory:
        error = "Not enough memory";
        break;
    }

    if (warning)
        printf("%s Xpm Warning: %s.\n", tag, warning);

    if (error) {
        printf("%s Xpm Error: %s.\n", tag, error);
/*      exit( -1 ); */
    }
}


void createMaterialPredefPanel()
{
    extern void create_menus();
    extern XtActionProc CanvasMaterialPredef(), CanvasMaterialPredefQuit();

    typedef struct _menu_struct {
        char                *name;           /* name of the btn */
        XtActionProc       (*func)();        /* callback to be invoked */
        char                *data;           /* data for the callback */
        struct _menu_struct *sub_menu;       /* data for submenu of this btn */
        int                  n_sub_items;    /* how many items in sub_menu */
        char                *sub_menu_title; /* title of submenu */
    } menu_struct;

    static menu_struct MetalData[] = {
        { "brass",          CanvasMaterialPredef,  "0", NULL, 0, NULL },
        { "bronze",         CanvasMaterialPredef,  "1", NULL, 0, NULL },
        { "bronzePolished", CanvasMaterialPredef,  "2", NULL, 0, NULL },
        { "chrome",         CanvasMaterialPredef,  "3", NULL, 0, NULL },
        { "copper",         CanvasMaterialPredef,  "4", NULL, 0, NULL },
        { "copperPolished", CanvasMaterialPredef,  "5", NULL, 0, NULL },
        { "gold",           CanvasMaterialPredef,  "5", NULL, 0, NULL },
        { "goldPolished",   CanvasMaterialPredef,  "7", NULL, 0, NULL },
        { "pewter",         CanvasMaterialPredef,  "8", NULL, 0, NULL },
        { "silver",         CanvasMaterialPredef,  "9", NULL, 0, NULL },
        { "silverPolished", CanvasMaterialPredef, "10", NULL, 0, NULL },
    };

    static menu_struct PlasticData[] = {
        { "blackPlastic",   CanvasMaterialPredef, "11", NULL, 0, NULL },
        { "bluePlastic",    CanvasMaterialPredef, "12", NULL, 0, NULL },
        { "cyanPlastic",    CanvasMaterialPredef, "13", NULL, 0, NULL },
        { "greenPlastic",   CanvasMaterialPredef, "14", NULL, 0, NULL },
        { "greyPlastic",    CanvasMaterialPredef, "15", NULL, 0, NULL },
        { "pinkPlastic",    CanvasMaterialPredef, "16", NULL, 0, NULL },
        { "redPlastic",     CanvasMaterialPredef, "17", NULL, 0, NULL },
        { "whitePlastic",   CanvasMaterialPredef, "18", NULL, 0, NULL },
        { "yellowPlastic",  CanvasMaterialPredef, "19", NULL, 0, NULL },
    };

    static menu_struct RubberData[] = {
        { "blackRubber",    CanvasMaterialPredef, "20", NULL, 0, NULL },
        { "blueRubber",     CanvasMaterialPredef, "23", NULL, 0, NULL },
        { "cyanRubber",     CanvasMaterialPredef, "26", NULL, 0, NULL },
        { "greenRubber",    CanvasMaterialPredef, "21", NULL, 0, NULL },
        { "magentaRubber",  CanvasMaterialPredef, "24", NULL, 0, NULL },
        { "orangeRubber",   CanvasMaterialPredef, "27", NULL, 0, NULL },
        { "redRubber",      CanvasMaterialPredef, "22", NULL, 0, NULL },
        { "whiteRubber",    CanvasMaterialPredef, "25", NULL, 0, NULL },
        { "yellowRubber",   CanvasMaterialPredef, "28", NULL, 0, NULL },
    };

    static menu_struct StoneData[] = {
        { "emerald",        CanvasMaterialPredef, "29", NULL, 0, NULL },
        { "jade",           CanvasMaterialPredef, "32", NULL, 0, NULL },
        { "obsidian",       CanvasMaterialPredef, "30", NULL, 0, NULL },
        { "pearl",          CanvasMaterialPredef, "33", NULL, 0, NULL },
        { "ruby",           CanvasMaterialPredef, "31", NULL, 0, NULL },
        { "turquoise",      CanvasMaterialPredef, "34", NULL, 0, NULL },
    };

    static menu_struct OtherData[] = {
        { "default",        CanvasMaterialPredef, "35", NULL, 0, NULL },
        { "brown",          CanvasMaterialPredef, "36", NULL, 0, NULL },
        { "darkbrown",      CanvasMaterialPredef, "37", NULL, 0, NULL },
        { "gray",           CanvasMaterialPredef, "38", NULL, 0, NULL },
        { "glass",          CanvasMaterialPredef, "39", NULL, 0, NULL },
        { "green",          CanvasMaterialPredef, "40", NULL, 0, NULL },
        { "red",            CanvasMaterialPredef, "41", NULL, 0, NULL },
        { "sea",            CanvasMaterialPredef, "42", NULL, 0, NULL },
        { "white",          CanvasMaterialPredef, "43", NULL, 0, NULL },
    };

    /*
     * Describe the menu bar, giving only the names to appear in
     * the menu bar and pointers to each pulldown pane.
     */
    static menu_struct PulldownData[] = {
        { "Metal",   NULL, NULL, MetalData,   XtNumber(MetalData),   NULL },
        { "Plastic", NULL, NULL, PlasticData, XtNumber(PlasticData), NULL },
        { "Rubber",  NULL, NULL, RubberData,  XtNumber(RubberData),  NULL },
        { "Stone",   NULL, NULL, StoneData,   XtNumber(StoneData),   NULL },
        { "Other",   NULL, NULL, OtherData,   XtNumber(OtherData),   NULL },
    };

    Position x, y;

    Widget w_menu, w_f, w_quit;

#ifdef XPM
    int status;
    XpmAttributes attributes;
    Pixmap grayPixmap, pix;
#endif

    map_dialog( w_material_predef, &x, &y );
    w_control_material_predef = XtVaCreatePopupShell( "Predef Material Panel",
        transientShellWidgetClass,
        w_material_predef,
        XmNx, x,
        XmNy, y,
        NULL );

    w_f = XtVaCreateManagedWidget( "w_f", xmFormWidgetClass,
        w_control_material_predef,
#ifdef XPM
        XmNbackgroundPixmap, groundPixmap,
#endif        /*  */
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNrightAttachment, XmATTACH_FORM,
        XmNrightOffset, 5,
        XmNtopAttachment, XmATTACH_FORM,
        XmNtopOffset, 5,
        XmNbottomAttachment, XmATTACH_FORM,
        XmNbottomOffset, 5,
        XmNwidth, 310,
        XmNheight, 30,
        NULL );

    w_menu = XmCreateMenuBar( w_f, "w_menu", NULL, 0 );

    XtVaSetValues( w_menu,
#ifdef XPM
        XmNbackgroundPixmap, groundPixmap,
#endif        /*  */
        XmNshadowThickness, 0,
        XmNhighlightThickness, 0,
        NULL );
    XtManageChild( w_menu );

    create_menus( NULL, w_menu, PulldownData, XtNumber( PulldownData ),
        groundPixmap );

    w_quit = XtVaCreateManagedWidget( "w_quit", xmPushButtonWidgetClass,
        w_f,
        XmNlabelString, XmStringCreateSimple( "Quit" ),
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_menu,
        XmNleftOffset, 0,
        XmNtopAttachment, XmATTACH_FORM,
        XmNtopOffset, 3,
        XmNrightAttachment, XmATTACH_FORM,
        XmNrightOffset, 3,
        XmNwidth, strlen("Quit") * 9,
        XmNheight, 22,
        NULL );
    XtAddCallback( w_quit, XmNactivateCallback,
        (XtCallbackProc) CanvasMaterialPredefQuit, (XtPointer)NULL );

    XtRealizeWidget( w_control_material_predef );
    XtPopup( w_control_material_predef, XtGrabNone );
}


void createVideoPanel()
{
    Position x, y;

    extern XtActionProc CanvasSavePPM(), CanvasSaveMPG(),
        CanvasStepBack(), CanvasPlay(), CanvasStepForw(), CanvasStop(),
        CanvasRecord(), CanvasVideoQuit();

    Widget w_vcr, w_sback, w_play, w_sforw, w_stop,
        w_record, w_savePPM, w_saveMPG,
        w_quit;

#ifdef XPM
    int status;
    XpmAttributes attributes;
    Pixmap recPixmap, grayPixmap, pix;
    Pixmap sbackPixmap, playPixmap, sforwPixmap, stopPixmap;
#endif

    map_dialog( w_video, &x, &y );
    w_control_video = XtVaCreatePopupShell( "Video Panel",
        transientShellWidgetClass,
        w_video,
        XmNx, x,
        XmNy, y,
        NULL );

    w_vcr = XtVaCreateManagedWidget( "w_vcr", xmFormWidgetClass,
        w_control_video,
#ifdef XPM
        XmNbackgroundPixmap, groundPixmap,
#endif        /*  */
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNrightAttachment, XmATTACH_FORM,
        XmNrightOffset, 5,
        XmNtopAttachment, XmATTACH_FORM,
        XmNtopOffset, 5,
        XmNbottomAttachment, XmATTACH_FORM,
        XmNbottomOffset, 5,
        XmNwidth, 130,
        XmNheight, 4 * 32,
        NULL );

#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/gray.xpm", &grayPixmap, &pix, &attributes );
    errorMessageXPM( status, "Gray" );
#endif        /*  */
#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/sback.xpm", &sbackPixmap, &pix, &attributes );
    errorMessageXPM( status, "Back" );
#endif        /*  */
    w_sback = XtVaCreateManagedWidget( "w_sback", xmPushButtonWidgetClass,
        w_vcr,
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_FORM,
        XmNtopOffset, 5,
        XmNwidth, 25,
        XmNheight, 25,
#ifdef XPM
        XmNlabelType, XmPIXMAP,
        XmNlabelPixmap, sbackPixmap,
        XmNbackgroundPixmap, grayPixmap,
#else
        XmNlabelString, XmStringCreateSimple( "Back" ),
#endif        /*  */
        NULL );
    XtAddCallback( w_sback, XmNactivateCallback, (XtCallbackProc)CanvasStepBack,
        (XtPointer)NULL );

#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/play.xpm", &playPixmap, &pix, &attributes );
    errorMessageXPM( status, "Play" );
#endif        /*  */
    w_play = XtVaCreateManagedWidget( "w_play", xmPushButtonWidgetClass,
        w_vcr,
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_sback,
        XmNleftOffset, 6,
        XmNtopAttachment, XmATTACH_FORM,
        XmNtopOffset, 5,
        XmNwidth, 25,
        XmNheight, 25,
#ifdef XPM
        XmNlabelType, XmPIXMAP,
        XmNlabelPixmap, playPixmap,
        XmNbackgroundPixmap, grayPixmap,
#else
        XmNlabelString, XmStringCreateSimple( "Play" ),
#endif        /*  */
        NULL );
    XtAddCallback( w_play, XmNactivateCallback, (XtCallbackProc)CanvasPlay,
        (XtPointer)NULL );

#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/sforw.xpm", &sforwPixmap, &pix, &attributes );
    errorMessageXPM( status, "Forw" );
#endif        /*  */
    w_sforw = XtVaCreateManagedWidget( "w_sforw", xmPushButtonWidgetClass,
        w_vcr,
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_play,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_FORM,
        XmNtopOffset, 5,
        XmNwidth, 25,
        XmNheight, 25,
#ifdef XPM
        XmNlabelType, XmPIXMAP,
        XmNlabelPixmap, sforwPixmap,
        XmNbackgroundPixmap, grayPixmap,
#else
        XmNlabelString, XmStringCreateSimple( "Forw" ),
#endif        /*  */
        NULL );
    XtAddCallback( w_sforw, XmNactivateCallback, (XtCallbackProc)CanvasStepForw,
        (XtPointer)NULL );

#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/stop.xpm", &stopPixmap, &pix, &attributes );
    errorMessageXPM( status, "Stop" );
#endif        /*  */
    w_stop = XtVaCreateManagedWidget( "w_stop", xmPushButtonWidgetClass,
        w_vcr,
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_sforw,
        XmNleftOffset, 6,
        XmNtopAttachment, XmATTACH_FORM,
        XmNtopOffset, 5,
        XmNwidth, 25,
        XmNheight, 25,
#ifdef XPM
        XmNlabelType, XmPIXMAP,
        XmNlabelPixmap, stopPixmap,
        XmNbackgroundPixmap, grayPixmap,
#else
        XmNlabelString, XmStringCreateSimple( "Stop" ),
#endif        /*  */
        NULL );
    XtAddCallback( w_stop, XmNactivateCallback, (XtCallbackProc)CanvasStop,
        (XtPointer)NULL );

#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/rec.xpm", &recPixmap, &pix, &attributes );
    errorMessageXPM( status, "Record" );
#endif        /*  */
    w_record = XtVaCreateManagedWidget( "w_record", xmPushButtonWidgetClass,
        w_vcr,
        XmNlabelString, XmStringCreateSimple( " REC" ),
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_play,
        XmNtopOffset, 5,
        XmNwidth, strlen(" REC ") * 9,
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, recPixmap,
#endif        /*  */
        NULL );
    XtAddCallback( w_record, XmNactivateCallback, (XtCallbackProc)CanvasRecord,
        (XtPointer)NULL );

    w_savePPM = XtVaCreateManagedWidget( "w_savePPM", xmPushButtonWidgetClass,
        w_vcr,
        XmNlabelString, XmStringCreateSimple( "PPM" ),
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_record,
        XmNtopOffset, 5,
        XmNwidth, (int)((int)(strlen("PlayStopREC  ") * 9) / (int)2 - 2),
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        NULL );
    XtAddCallback( w_savePPM, XmNactivateCallback,(XtCallbackProc)CanvasSavePPM,
        (XtPointer)NULL );

    w_saveMPG = XtVaCreateManagedWidget( "w_saveMPG", xmPushButtonWidgetClass,
        w_vcr,
        XmNlabelString, XmStringCreateSimple( "MPG" ),
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_savePPM,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_record,
        XmNtopOffset, 5,
        XmNwidth, (int)((int)(strlen("PlayStopREC  ") * 9) / (int)2 - 2),
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        NULL );
    XtAddCallback( w_saveMPG, XmNactivateCallback,(XtCallbackProc)CanvasSaveMPG,
        (XtPointer)NULL );

    w_quit = XtVaCreateManagedWidget( "w_quit", xmPushButtonWidgetClass,
        w_vcr,
        XmNlabelString, XmStringCreateSimple( "Quit" ),
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_saveMPG,
        XmNtopOffset, 5,
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNbottomAttachment, XmATTACH_FORM,
        XmNbottomOffset, 5,
        XmNwidth, strlen("Quit") * 9,
        XmNheight, 25,
        NULL );
    XtAddCallback( w_quit, XmNactivateCallback, (XtCallbackProc)CanvasVideoQuit,
        (XtPointer)NULL );

    XtRealizeWidget( w_control_video );
    XtPopup( w_control_video, XtGrabNone );
}


void createControlPanel()
{
    Widget w, w_box;
    Widget w_reset,
           w_color, w_material, w_light,
           w_flyby, w_animcr, w_animex, w_morph, w_spin,
           w_head, w_headUp, w_headDo, w_headLe, w_headRi,
           w_hardcopy, w_quit, w_pex;
    Position x, y;

    extern int no_move;

#ifdef XPM
    int status;
    XpmAttributes attributes;
    Pixmap wirePixmap, colorPixmap, flybyPixmap, grayPixmap;
    Pixmap pexPixmap, pix;
    Pixmap upPixmap, doPixmap, lePixmap, riPixmap;
#endif

    extern int object_defined;
    extern XtActionProc CanvasReset(), CanvasQuit(),
        CanvasWireframe(), CanvasCull(), CanvasEdge(), CanvasShadow(),
        CanvasColor(), CanvasMaterial(), CanvasLight(),
        CanvasFlyby(), CanvasAnimateCr(), CanvasAnimateEx(),
        CanvasMorph(), CanvasSpin(),
        CanvasHardcopy(),
        CanvasUp(), CanvasDown(), CanvasLeft(), CanvasRight();

    static XtCallbackRec callbacks[] =
    {
        {(XtCallbackProc) NULL, (XtPointer) NULL},
        {(XtCallbackProc) NULL, (XtPointer) NULL},
    };

    if( !object_defined )
        map_dialog( toplevel, &x, &y );
    w_control = XtVaCreatePopupShell( "Control Panel",
        xmDialogShellWidgetClass,
        toplevel,
        XmNautoUnmanage, False,
        XmNdefaultPosition, False,
        XmNx, ((x+650) > WidthOfScreen( XtScreen( toplevel )) ? x-350 : x+650),
        XmNy, y-50,
        NULL );

#if( 0 )
    XtVaSetValues( w_control,
        XmNmwmDecorations, MWM_DECOR_BORDER | MWM_DECOR_TITLE |
            MWM_DECOR_MINIMIZE,
        XmNmwmFunctions, MWM_FUNC_MOVE | MWM_FUNC_MINIMIZE | MWM_FUNC_CLOSE,
            NULL );
#endif

#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/background.xpm", &groundPixmap, &pix, &attributes );
    errorMessageXPM( status, "Background" );
#endif        /*  */

    w = XtVaCreateManagedWidget( "w_box", xmRowColumnWidgetClass, w_control,
#ifdef XPM
        XmNbackgroundPixmap, groundPixmap,
#endif        /*  */
        XmNwidth, 215,
        XmNheight, 11 * 31,
        NULL );

    w_box = XtVaCreateManagedWidget( "w_box", xmFormWidgetClass, w,
#ifdef XPM
        XmNbackgroundPixmap, groundPixmap,
#endif        /*  */
        NULL );

#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/gray.xpm", &grayPixmap, &pix, &attributes );
    errorMessageXPM( status, "Gray" );
#endif        /*  */
    w_reset = XtVaCreateManagedWidget( "w_reset", xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Reset" ),
        XmNtopAttachment, XmATTACH_FORM,
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNtopOffset, 5,
        XmNwidth, strlen("Reset") * 9,
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        NULL );
    XtAddCallback( w_reset, XmNactivateCallback, (XtCallbackProc)CanvasReset,
        (XtPointer)NULL );

#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/wire.xpm", &wirePixmap, &pix, &attributes );
    errorMessageXPM( status, "Wire" );
#endif        /*  */
    w_wireframe = XtVaCreateManagedWidget( "w_wireframe",
        xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Solid" ),
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_reset,
        XmNtopOffset, 5,
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNrecomputeSize, False,
        XmNwidth, strlen("Solid ") * 9,
        XmNheight, 25,
#ifdef XPM
        /* XmNbackgroundPixmap, wirePixmap, */
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        NULL );
    XtAddCallback( w_wireframe, XmNactivateCallback,
        (XtCallbackProc)CanvasWireframe, (XtPointer)NULL );

    w_cull = XtVaCreateManagedWidget( "w_cull",
        xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Cull On" ),
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_reset,
        XmNtopOffset, 5,
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_wireframe,
        XmNleftOffset, 5,
        XmNrecomputeSize, False,
        XmNwidth, strlen("Cull Off") * 9,
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        NULL );
    XtAddCallback( w_cull, XmNactivateCallback, (XtCallbackProc)CanvasCull,
        (XtPointer)NULL );

    w_edge = XtVaCreateManagedWidget( "w_edge",
        xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Edges Off" ),
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_reset,
        XmNtopOffset, 5,
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_cull,
        XmNleftOffset, 5,
        XmNrecomputeSize, False,
        XmNwidth, strlen("Edges Off") * 8,
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        NULL );
    XtAddCallback( w_edge, XmNactivateCallback, (XtCallbackProc)CanvasEdge,
        (XtPointer)NULL );

    w_shadow = XtVaCreateManagedWidget( "w_shadow",
        xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Shadow" ),
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_wireframe,
        XmNtopOffset, 5,
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNwidth, strlen("Shadow") * 9,
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        NULL );
    XtAddCallback( w_shadow, XmNactivateCallback, (XtCallbackProc)CanvasShadow,
        (XtPointer)NULL );

    w_morph = XtVaCreateManagedWidget( "w_morph", xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Morph" ),
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_shadow,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_wireframe,
        XmNtopOffset, 5,
        XmNwidth, strlen("Morph ") * 9,
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        NULL );
    XtAddCallback( w_morph, XmNactivateCallback, (XtCallbackProc)CanvasMorph,
        (XtPointer)NULL );

    w_spin = XtVaCreateManagedWidget( "w_spin", xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Auto Spin" ),
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_morph,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_wireframe,
        XmNtopOffset, 5,
        XmNwidth, strlen("AutoSpin") * 9,
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        NULL );
    XtAddCallback( w_spin, XmNactivateCallback, (XtCallbackProc)CanvasSpin,
        (XtPointer)NULL );

#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/color.xpm", &colorPixmap, &pix, &attributes );
    errorMessageXPM( status, "Color" );
#endif        /*  */
    w_color = XtVaCreateManagedWidget( "w_color", xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Set Color" ),
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_shadow,
        XmNtopOffset, 5,
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNwidth, strlen("SetColor") * 9,
        XmNheight, 25,
#ifdef XPM
        /* XmNbackgroundPixmap, colorPixmap, */
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        NULL );
    XtAddCallback( w_color, XmNactivateCallback, (XtCallbackProc)CanvasColor,
        (XtPointer)NULL );

    w_light = XtVaCreateManagedWidget( "w_light", xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Set Light" ),
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_shadow,
        XmNtopOffset, 5,
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_color,
        XmNleftOffset, 5,
        XmNwidth, strlen("SetLight") * 9,
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        NULL );
    XtAddCallback( w_light, XmNactivateCallback, (XtCallbackProc)CanvasLight,
        (XtPointer)NULL );

    w_material = XtVaCreateManagedWidget( "w_material", xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Set Material" ),
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_color,
        XmNtopOffset, 5,
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
/*      XmNwidth, strlen("SetMaterial") * 9, */
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        NULL );
    XtAddCallback( w_material, XmNactivateCallback,
        (XtCallbackProc)CanvasMaterial, (XtPointer)NULL );

    callbacks[0].callback = (XtCallbackProc) createMaterialPredefPanel;
    callbacks[0].closure  = (XtPointer) NULL;
    w_material_predef = XtVaCreateManagedWidget( "w_material_predef",
        xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Predef Material" ),
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_color,
        XmNtopOffset, 5,
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_material,
        XmNleftOffset, 5,
/*      XmNwidth, strlen("PredefMaterial") * 9, */
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        XmNactivateCallback, callbacks,
        NULL );

#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/flyby.xpm", &flybyPixmap, &pix, &attributes );
    errorMessageXPM( status, "Flyby" );
#endif        /*  */
    w_flyby = XtVaCreateManagedWidget( "w_flyby", xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Fly by" ),
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_material,
        XmNtopOffset, 5,
        XmNwidth, strlen("Fly by") * 9,
        XmNheight, 25,
#ifdef XPM
        /* XmNbackgroundPixmap, flybyPixmap, */
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        NULL );
    XtAddCallback( w_flyby, XmNactivateCallback, (XtCallbackProc)CanvasFlyby,
        (XtPointer)NULL );

    callbacks[0].callback = (XtCallbackProc) CanvasAnimateCr;
    callbacks[0].closure  = (XtPointer) NULL;
    w_animcr = XtVaCreateManagedWidget( "w_animcr", xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Anim Create" ),
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_flyby,
        XmNtopOffset, 5,
        XmNwidth, strlen("AnimCreate") * 8,
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        XmNactivateCallback, callbacks,
        NULL );

    callbacks[0].callback = (XtCallbackProc) CanvasAnimateEx;
    callbacks[0].closure  = (XtPointer) NULL;
    w_animex = XtVaCreateManagedWidget( "w_animex", xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Anim Execute" ),
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_animcr,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_flyby,
        XmNtopOffset, 5,
        XmNwidth, strlen("AnimExecute") * 8,
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        XmNactivateCallback, callbacks,
        NULL );
#if( 0 )
    w_head = XtVaCreateManagedWidget( "w_head", xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Head:" ),
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_flyby,
        XmNtopOffset, 5,
        XmNwidth, strlen("Head:") * 9,
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        NULL );

#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/left.xpm", &lePixmap, &pix, &attributes );
    errorMessageXPM( status, "Left" );
#endif        /*  */
    w_headLe = XtVaCreateManagedWidget( "w_headLe", xmPushButtonWidgetClass,
        w_box,
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_head,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_head,
        XmNtopOffset, 5,
#ifdef XPM
        XmNlabelType, XmPIXMAP,
        XmNlabelPixmap, lePixmap,
        XmNbackgroundPixmap, lePixmap,
#else
        XmNlabelString, XmStringCreateSimple( "Le" ),
#endif        /*  */
        XmNwidth, 25,
        XmNheight, 25,
        NULL );
    XtAddCallback( w_headLe, XmNactivateCallback, (XtCallbackProc)CanvasLeft,
        (XtPointer)NULL );

#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/up.xpm", &upPixmap, &pix, &attributes );
    errorMessageXPM( status, "Up" );
#endif        /*  */
    w_headUp = XtVaCreateManagedWidget( "w_headUp", xmPushButtonWidgetClass,
        w_box,
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_headLe,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_flyby,
        XmNtopOffset, 5,
#ifdef XPM
        XmNlabelType, XmPIXMAP,
        XmNlabelPixmap, upPixmap,
        XmNbackgroundPixmap, upPixmap,
#else
        XmNlabelString, XmStringCreateSimple( "Up" ),
#endif        /*  */
        XmNwidth, 25,
        XmNheight, 25,
        NULL );
    XtAddCallback( w_headUp, XmNactivateCallback, (XtCallbackProc)CanvasUp,
        (XtPointer)NULL );

#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/down.xpm", &doPixmap, &pix, &attributes );
    errorMessageXPM( status, "Down" );
#endif        /*  */
    w_headDo = XtVaCreateManagedWidget( "w_headDo", xmPushButtonWidgetClass,
        w_box,
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_headLe,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_headLe,
        XmNtopOffset, 5,
#ifdef XPM
        XmNlabelType, XmPIXMAP,
        XmNlabelPixmap, doPixmap,
        XmNbackgroundPixmap, doPixmap,
#else
        XmNlabelString, XmStringCreateSimple( "Do" ),
#endif        /*  */
        XmNwidth, 25,
        XmNheight, 25,
        NULL );
    XtAddCallback( w_headDo, XmNactivateCallback, (XtCallbackProc)CanvasDown,
        (XtPointer)NULL );

#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/right.xpm", &riPixmap, &pix, &attributes );
    errorMessageXPM( status, "Right" );
#endif        /*  */
    w_headRi = XtVaCreateManagedWidget( "w_headRi", xmPushButtonWidgetClass,
        w_box,
        XmNleftAttachment, XmATTACH_WIDGET,
        XmNleftWidget, w_headDo,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_headUp,
        XmNtopOffset, 5,
#ifdef XPM
        XmNlabelType, XmPIXMAP,
        XmNlabelPixmap, riPixmap,
        XmNbackgroundPixmap, riPixmap,
#else
        XmNlabelString, XmStringCreateSimple( "Ri" ),
#endif        /*  */
        XmNwidth, 25,
        XmNheight, 25,
        NULL );
    XtAddCallback( w_headRi, XmNactivateCallback, (XtCallbackProc)CanvasRight,
        (XtPointer)NULL );
#endif        /*  */

    callbacks[0].callback = (XtCallbackProc) createVideoPanel;
    callbacks[0].closure  = (XtPointer) NULL;
    w_video = XtVaCreateManagedWidget( "w_video", xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Video" ),
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_animcr,
        XmNtopOffset, 5,
        XmNwidth, strlen("Video") * 9,
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        XmNactivateCallback, callbacks,
        NULL );

    w_hardcopy = XtVaCreateManagedWidget( "w_hardcopy", xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Hardcopy" ),
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_video,
        XmNtopOffset, 5,
        XmNwidth, strlen("Hardcopy") * 9,
        XmNheight, 25,
#ifdef XPM
        XmNbackgroundPixmap, grayPixmap,
#endif        /*  */
        NULL );
    XtAddCallback( w_hardcopy, XmNactivateCallback,
        (XtCallbackProc)CanvasHardcopy, (XtPointer)NULL );

    w_quit = XtVaCreateManagedWidget( "w_quit", xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( "Quit" ),
        XmNtopAttachment, XmATTACH_WIDGET,
        XmNtopWidget, w_hardcopy,
        XmNtopOffset, 10,
        XmNleftAttachment, XmATTACH_FORM,
        XmNleftOffset, 5,
        XmNbottomAttachment, XmATTACH_FORM,
        XmNbottomOffset, 10,
        XmNwidth, strlen("Quit") * 9,
        XmNheight, strlen("Quit") * 9,
        NULL );
    XtAddCallback( w_quit, XmNactivateCallback, (XtCallbackProc)CanvasQuit,
        (XtPointer)NULL );

#ifdef XPM
    attributes.valuemask = XpmReturnPixels;
    status = XpmReadFileToPixmap( display,
        DefaultRootWindow(XtDisplay(toplevel)),
        "pixmaps/pex.xpm", &pexPixmap, &pix, &attributes );
    errorMessageXPM( status, "Pex" );
#endif        /*  */
    w_pex = XtVaCreateManagedWidget( "w_pex", xmPushButtonWidgetClass,
        w_box,
        XmNlabelString, XmStringCreateSimple( " " ),
        XmNrightAttachment, XmATTACH_FORM,
        XmNrightOffset, 10,
        XmNbottomAttachment, XmATTACH_FORM,
        XmNbottomOffset, 10,
#ifdef XPM
        XmNlabelType, XmPIXMAP,
        XmNlabelPixmap, pexPixmap,
        XmNbackgroundPixmap, pexPixmap,
#endif        /*  */
        XmNshadowThickness, 0,
        XmNhighlightThickness, 0,
        XmNwidth, 42,
        XmNheight, 42,
        NULL );

    XtManageChild( w_control );
}
